#ifdef _WIN32
#include <winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <unistd.h>
#include <arpa/inet.h>
#endif

#include <sys/time.h>
#include <sys/types.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>

#include <iomanip>
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>

#include <libssh2.h>
#include <libssh2_sftp.h>


const int SSH_EAGAIN_TIMEOUT_MS = 5000;
const char *hostname = "192.168.1.20";
const char *username = "ubnt";
const char *password = "ubnt";


bool bFirstTime = true;
std::ofstream sshLogFile;
void libssh2_trace_handler_cb(LIBSSH2_SESSION *session, void *opaque, const char *msg, size_t msgSz)
{
    if (bFirstTime)
    {
        sshLogFile.open("ssh.log");
        bFirstTime = false;
    }
    sshLogFile << msg << std::endl;
}


void ncvDebugOutput(const std::string &msg)
{
    if (bFirstTime)
    {
        sshLogFile.open("ssh.log");
        bFirstTime = false;
    }
    sshLogFile << msg;
    std::cout << msg;
}


#define ncvPrint(msg) \
    do \
    { \
        std::ostringstream oss; \
        oss << msg << std::endl; \
        ncvDebugOutput(oss.str()); \
    } while (0)


#define ncvAssertPrint(pred, msg, isWarning) \
    do \
    { \
        if (!(pred)) \
        { \
            std::ostringstream oss; \
            oss << "### " << (isWarning ? "Warning: " : "Error: ") << msg << \
                   "\n    file=" << __FILE__ << \
                   "\n    line=" << __LINE__ << \
                   std::endl; \
            ncvDebugOutput(oss.str()); \
        } \
    } while (0)


#define ncvAssertReturn(pred, msg, retCode) \
    do \
    { \
        bool _pred = (pred); \
        ncvAssertPrint(_pred, msg, false); \
        if (!_pred) \
        { \
            return retCode; \
        } \
    } while (0)


class AutoWsaInit
{
public:
    AutoWsaInit(bool &bInit)
    {
#ifdef _WIN32
        this->bInit = false;
        bInit = false;
        WSADATA wsaData;
        int res = WSAStartup(MAKEWORD(1, 1), &wsaData);
        ncvAssertReturn(res == 0, "WSAStartup failed", );
        this->bInit = true;
#endif
        bInit = true;
    }

    ~AutoWsaInit()
    {
#ifdef _WIN32
        if (this->bInit)
        {
            WSACleanup();
        }
#endif
    }

private:
    bool bInit;
};


class AutoLibssh2Init
{
public:
    AutoLibssh2Init(bool &bInit)
    {
        this->bInit = false;
        bInit = false;

        int res = libssh2_init(0);
        ncvAssertReturn(res == 0, "LibSSH2 initialization failure, res=" << res, );

        this->bInit = true;
        bInit = true;
    }

    ~AutoLibssh2Init()
    {
        if (this->bInit)
        {
            libssh2_exit();
        }
    }
private:
    bool bInit;
};


class AutoSocket
{
public:
    AutoSocket(const std::string &ipAddr, unsigned ipPort, libssh2_socket_t &sock, bool &bInit)
    {
        int res;
        this->bInit = false;
        bInit = false;

        this->sock = socket(AF_INET, SOCK_STREAM, 0);
        ncvAssertReturn(this->sock != 0, "Socket create failure", );

        unsigned hostaddr = inet_addr(ipAddr.c_str());

        sockaddr_in sockAddrIn;
        sockAddrIn.sin_family = AF_INET;
        sockAddrIn.sin_port = htons(ipPort);
        sockAddrIn.sin_addr.s_addr = hostaddr;
        res = connect(this->sock, (sockaddr*)(&sockAddrIn), sizeof(sockaddr_in));
        ncvAssertReturn(res == 0, "Socket connect failure, res=" << res, );

        sock = this->sock;
        this->bInit = true;
        bInit = true;
    }

    ~AutoSocket()
    {
        if (this->bInit)
        {
#ifdef WIN32
            closesocket(sock);
#else
            close(sock);
#endif
        }
    }

private:
    bool bInit;
    libssh2_socket_t sock;
};


struct LibSsh2SessionWrap
{
    LIBSSH2_SESSION *session;
    libssh2_socket_t sock;
};


static int timedWaitSessionSocket(LIBSSH2_SESSION *session, libssh2_socket_t sock, int timeoutMs)
{
    fd_set fd;
    FD_ZERO(&fd);
    FD_SET(sock, &fd);

    int dir = libssh2_session_block_directions(session);

    fd_set *readfd = NULL;
    if (dir & LIBSSH2_SESSION_BLOCK_INBOUND)
    {
        readfd = &fd;
    }

    fd_set *writefd = NULL;
    if (dir & LIBSSH2_SESSION_BLOCK_OUTBOUND)
    {
        writefd = &fd;
    }

    timeval timeout;
    timeout.tv_sec = timeoutMs / 1000;
    timeout.tv_usec = (timeoutMs % 1000) * 1000;
    int res = select(sock+1, readfd, writefd, NULL, &timeout);
    if (res == 0)
    {
        ncvAssertReturn(false, "select timeout", LIBSSH2_ERROR_EAGAIN);
    }
    else if (res < 0)
    {
        ncvAssertReturn(false, "select error=" << res, -1);
    }
    return 0;
}


#define LIBSSH2_TIMED_PREFIX(libssh2_func_name) \
    ncvPrint(">>>>> Begin " libssh2_func_name); \
    int res; \
    while (1) \
    {

#define LIBSSH2_TIMED_SUFFIX_INT(libssh2_func_name, BWAITSOCKET, SESSION, SOCK, SFTP) \
        if (res != LIBSSH2_ERROR_EAGAIN) \
        { \
            if (SFTP != NULL && res == LIBSSH2_ERROR_SFTP_PROTOCOL) \
            { \
                res = libssh2_sftp_last_error(SFTP); \
            } \
            ncvPrint(">>>>> End " libssh2_func_name "res=" << res); \
            return res; \
        } \
        if (BWAITSOCKET) \
        { \
            res = timedWaitSessionSocket(SESSION, SOCK, timeoutMs); \
            ncvAssertReturn(res == 0, ">>>>> End " libssh2_func_name ": timedWaitSessionSocket failure, res=" << res, res); \
        } \
    }

#define LIBSSH2_TIMED_SUFFIX_CHANNEL(libssh2_func_name, chname, SESSION, SOCK, SFTP) \
        if (chname != NULL) \
        { \
            ncvPrint(">>>>> End " libssh2_func_name); \
            return 0; \
        } \
        res = libssh2_session_last_errno(SESSION); \
        if (res != LIBSSH2_ERROR_EAGAIN) \
        { \
            if (SFTP != NULL && res == LIBSSH2_ERROR_SFTP_PROTOCOL) \
            { \
                res = libssh2_sftp_last_error(SFTP); \
            } \
            ncvPrint(">>>>> End " libssh2_func_name "res=" << res); \
            return res; \
        } \
        res = timedWaitSessionSocket(SESSION, SOCK, timeoutMs); \
        ncvAssertReturn(res == 0, ">>>>> End " libssh2_func_name ": timedWaitSessionSocket failure, res=" << res, res); \
    }


static int timed_libssh2_session_handshake(LIBSSH2_SESSION *session, libssh2_socket_t sock, int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_session_handshake");
    res = libssh2_session_handshake(session, sock);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_session_handshake", 0, NULL, 0, NULL)
}


static int timed_libssh2_session_disconnect(LIBSSH2_SESSION *session,
                                            const char *description,
                                            int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_session_disconnect");
    res = libssh2_session_disconnect(session, description);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_session_disconnect", 0, NULL, 0, NULL)
}


static int timed_libssh2_session_free(LIBSSH2_SESSION *session, int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_session_free");
    res = libssh2_session_free(session);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_session_free", 0, NULL, 0, NULL)
}


static int timed_libssh2_userauth_password(LIBSSH2_SESSION *session,
                                           const char *username,
                                           const char *password,
                                           int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_userauth_password");
    res = libssh2_userauth_password(session, username, password);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_userauth_password", 0, NULL, 0, NULL)
}


static int timed_libssh2_sftp_init(const LibSsh2SessionWrap &session,
                                   LIBSSH2_SFTP * & sftpChannel,
                                   int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_init");
    sftpChannel = libssh2_sftp_init(session.session);
    LIBSSH2_TIMED_SUFFIX_CHANNEL("libssh2_sftp_init", sftpChannel, session.session, session.sock, NULL)
}


static int timed_libssh2_sftp_shutdown(const LibSsh2SessionWrap &session,
                                       LIBSSH2_SFTP *sftpChannel,
                                       int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_shutdown");
    res = libssh2_sftp_shutdown(sftpChannel);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_sftp_shutdown", 1, session.session, session.sock, sftpChannel)
}


static int timed_libssh2_sftp_open_ex2(const LibSsh2SessionWrap &session,
                                       LIBSSH2_SFTP *sftpChannel,
                                       const char *filename,
                                       unsigned flags,
                                       unsigned mode,
                                       LIBSSH2_SFTP_HANDLE * & sftpHandle,
                                       int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_open_ex");
    sftpHandle = libssh2_sftp_open_ex(sftpChannel, filename, strlen(filename), flags, mode, LIBSSH2_SFTP_OPENFILE);
    LIBSSH2_TIMED_SUFFIX_CHANNEL("libssh2_sftp_open_ex", sftpHandle, session.session, session.sock, sftpChannel)
}


static int timed_libssh2_sftp_close_handle(const LibSsh2SessionWrap &session,
                                           LIBSSH2_SFTP *sftpChannel,
                                           LIBSSH2_SFTP_HANDLE *sftpHandle,
                                           int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_close_handle");
    res = libssh2_sftp_close_handle(sftpHandle);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_sftp_close_handle", 1, session.session, session.sock, sftpChannel)
}


static int timed_libssh2_sftp_read(const LibSsh2SessionWrap &session,
                                   LIBSSH2_SFTP *sftpChannel,
                                   LIBSSH2_SFTP_HANDLE *sftpHandle,
                                   char *buffer,
                                   size_t buffer_maxlen,
                                   int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_read");
    res = libssh2_sftp_read(sftpHandle, buffer, buffer_maxlen);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_sftp_read", 1, session.session, session.sock, sftpChannel)
}


static int timed_libssh2_sftp_read_all(const LibSsh2SessionWrap &session,
                                       LIBSSH2_SFTP *sftpChannel,
                                       LIBSSH2_SFTP_HANDLE *sftpHandle,
                                       std::vector<unsigned char> &buf,
                                       int szMax,
                                       int timeoutMs)
{
    int res;
    buf.resize(0);
    char buffer[0x4000];
    int szLeft = (szMax > 0 ? szMax : INT_MAX);

    do
    {
        res = timed_libssh2_sftp_read(session, sftpChannel, sftpHandle, buffer, std::min<int>(szLeft, sizeof(buffer)), timeoutMs);
        ncvAssertReturn(res >= 0, "timed_libssh2_channel_read_ex failure, res=" << res, -1);

        if (res > 0)
        {
            unsigned oldSize = buf.size();
            buf.resize(oldSize + res);
            memcpy(&buf[0] + oldSize, buffer, res);
            if (szMax > 0 && (int)buf.size() >= szMax)
            {
                return 0;
            }
        }
    } while (res > 0);

    return 0;
}


static int timed_libssh2_sftp_write(const LibSsh2SessionWrap &session,
                                    LIBSSH2_SFTP *sftpChannel,
                                    LIBSSH2_SFTP_HANDLE *sftpHandle,
                                    const char *buffer,
                                    size_t count,
                                    int timeoutMs)
{
    LIBSSH2_TIMED_PREFIX("libssh2_sftp_write");
    res = libssh2_sftp_write(sftpHandle, buffer, count);
    LIBSSH2_TIMED_SUFFIX_INT("libssh2_sftp_write", 1, session.session, session.sock, sftpChannel)
}


static int timed_libssh2_sftp_write_all(const LibSsh2SessionWrap &session,
                                        LIBSSH2_SFTP *sftpChannel,
                                        LIBSSH2_SFTP_HANDLE *sftpHandle,
                                        const std::vector<unsigned char> &buf,
                                        int timeoutMs)
{
    int res;
    int bufOffs = 0;
    int bufLeft = (int)buf.size();

    do
    {
        res = timed_libssh2_sftp_write(session, sftpChannel, sftpHandle, (const char *)&buf[0]+bufOffs, bufLeft, timeoutMs);
        ncvAssertReturn(res >= 0, "timed_libssh2_sftp_write failure, res=" << res, -1);

        if (res > 0)
        {
            bufOffs += res;
            bufLeft -= res;
            if (bufLeft <= 0)
            {
                return 0;
            }
        }
    } while (res > 0);

    return 0;
}


class AutoSshSession
{
public:
    AutoSshSession(libssh2_socket_t sock, LIBSSH2_SESSION * & session, bool &bInit)
    {
        int res;
        this->bInit = false;
        bInit = false;
        session = NULL;

        this->session = libssh2_session_init();
        ncvAssertReturn(this->session != NULL, "libssh2_session_init failure", );

        libssh2_session_set_blocking(this->session, 0);

        res = timed_libssh2_session_handshake(this->session, sock, SSH_EAGAIN_TIMEOUT_MS);
        ncvAssertReturn(res == 0, "timed_libssh2_session_handshake failure, res=" << res, );

        this->bInit = true;
        bInit = true;
        session = this->session;

        const char *sessionHostkeyHash = libssh2_hostkey_hash(session, LIBSSH2_HOSTKEY_HASH_MD5);
        ncvAssertReturn(sessionHostkeyHash != NULL, "libssh2_hostkey_hash failure", );
    }

    int authenticate(const std::string &username, const std::string &password)
    {
        return timed_libssh2_userauth_password(this->session, username.c_str(), password.c_str(), SSH_EAGAIN_TIMEOUT_MS);
    }

    ~AutoSshSession()
    {
        if (this->bInit)
        {
            int res;
            res = timed_libssh2_session_disconnect(session, "bye", SSH_EAGAIN_TIMEOUT_MS);
            ncvAssertPrint(res == 0, "timed_libssh2_session_disconnect failure, res=" << res, false);
            res = timed_libssh2_session_free(session, SSH_EAGAIN_TIMEOUT_MS);
            ncvAssertPrint(res == 0, "timed_libssh2_session_free failure, res=" << res, false);
        }
    }

private:
    bool bInit;
    LIBSSH2_SESSION *session;
};


class AutoSftpChannel
{
public:
    AutoSftpChannel(const LibSsh2SessionWrap &session,
                    LIBSSH2_SFTP * & sftpChannel,
                    int &res)
      : session(session)
    {
        this->bInit = false;
        sftpChannel = NULL;
        res = timed_libssh2_sftp_init(session, sftpChannel, SSH_EAGAIN_TIMEOUT_MS);
        ncvAssertReturn(res == 0, "timed_libssh2_sftp_init failure, res=" << res, );
        this->sftpChannel = sftpChannel;
        this->bInit = true;
    }

    ~AutoSftpChannel()
    {
        if (this->bInit)
        {
            int res = timed_libssh2_sftp_shutdown(session, sftpChannel, SSH_EAGAIN_TIMEOUT_MS);
            ncvAssertPrint(res == 0, "timed_libssh2_sftp_shutdown failure, res=" << res, false);
        }
    }

private:
    bool bInit;
    const LibSsh2SessionWrap &session;
    LIBSSH2_SFTP *sftpChannel;
};


class AutoSftpHandle
{
public:
    AutoSftpHandle(const LibSsh2SessionWrap &session,
                   LIBSSH2_SFTP *sftpChannel,
                   const char *filename,
                   unsigned flags,
                   LIBSSH2_SFTP_HANDLE * & sftpHandle,
                   int timeoutMs,
                   int &res)
      : session(session),
        sftpChannel(sftpChannel),
        timeoutMs(timeoutMs)
    {
        this->bInit = false;
        sftpHandle = NULL;
        res = timed_libssh2_sftp_open_ex2(session, sftpChannel, filename, flags, 0666, sftpHandle, timeoutMs);
        ncvAssertReturn(res == 0, "timed_libssh2_sftp_open_ex2 failure, res=" << res, );
        this->sftpHandle = sftpHandle;
        this->bInit = true;
    }

    ~AutoSftpHandle()
    {
        if (this->bInit)
        {
            int res = timed_libssh2_sftp_close_handle(session, sftpChannel, sftpHandle, timeoutMs);
            ncvAssertPrint(res == 0, "timed_libssh2_sftp_close_handle failure, res=" << res, false);
        }
    }

private:
    bool bInit;
    const LibSsh2SessionWrap &session;
    LIBSSH2_SFTP *sftpChannel;
    int timeoutMs;
    LIBSSH2_SFTP_HANDLE *sftpHandle;
};


int main(int argc, char *argv[])
{
    int res;
    bool bInit;

    AutoWsaInit wsaInit(bInit);
    ncvAssertReturn(bInit, "AutoWsaInit ctor failure", -1);

    AutoLibssh2Init libssh2Init(bInit);
    ncvAssertReturn(bInit, "AutoLibssh2Init ctor failure", -1);

    libssh2_socket_t sock;
    AutoSocket aSocket(hostname, 22, sock, bInit);
    ncvAssertReturn(bInit, "AutoSocket ctor failure", -1);

    LIBSSH2_SESSION *session;
    AutoSshSession aSshSession(sock, session, bInit);
    ncvAssertReturn(bInit, "AutoSshSession ctor failure", -1);

    res = aSshSession.authenticate(username, password);
    ncvAssertReturn(res == 0, "SSH authentication failure", -1);

    libssh2_trace(session, ~0);
    libssh2_trace_sethandler(session, NULL, libssh2_trace_handler_cb);

    LibSsh2SessionWrap sessionWrap;
    sessionWrap.session = session;
    sessionWrap.sock = sock;

    std::vector<unsigned char> noise(1024*1024, 0);
    for (int i=0; i<(int)noise.size(); i++)
    {
        noise[i] = (unsigned char)(255 * ((double)rand()) / RAND_MAX);
    }

    //30000 is the first problematic size
    int sz[] = {/*1000, 2000, 3000, 4000, 5000, 10000, 20000, */30000, 40000, 50000, 80000, 100000, 200000, 400000, 800000, 1000000};

    LIBSSH2_SFTP *sftpChannel;
    AutoSftpChannel aSftpChannel(sessionWrap, sftpChannel, res);
    ncvAssertReturn(res == 0, "AutoSftpChannel ctor failure", -1);

    for (unsigned i=0; i<sizeof(sz)/sizeof(int); i++)
    {
        std::vector<unsigned char> subNoise(&noise[0], &noise[0] + sz[i]);

        LIBSSH2_SFTP_HANDLE *sftpHandleUp;
        AutoSftpHandle aSftpHandleUp(sessionWrap, sftpChannel, "/tmp/test", LIBSSH2_FXF_CREAT|LIBSSH2_FXF_TRUNC|LIBSSH2_FXF_WRITE, sftpHandleUp, SSH_EAGAIN_TIMEOUT_MS, res);
        ncvAssertReturn(res == 0, "AutoSftpHandle ctor failure", -1);

        res = timed_libssh2_sftp_write_all(sessionWrap, sftpChannel, sftpHandleUp, subNoise, SSH_EAGAIN_TIMEOUT_MS);
        ncvAssertReturn(res == 0, "timed_libssh2_sftp_write_all, res=" << res, -1);

        LIBSSH2_SFTP_HANDLE *sftpHandleDown;
        AutoSftpHandle aSftpHandleDown(sessionWrap, sftpChannel, "/tmp/test", LIBSSH2_FXF_READ, sftpHandleDown, SSH_EAGAIN_TIMEOUT_MS, res);
        ncvAssertReturn(res == 0, "AutoSftpHandle ctor failure", -1);

        std::vector<unsigned char> subNoiseRecvd;
        res = timed_libssh2_sftp_read_all(sessionWrap, sftpChannel, sftpHandleDown, subNoiseRecvd, 0, SSH_EAGAIN_TIMEOUT_MS);
        ncvAssertReturn(res == 0, "timed_libssh2_sftp_write_all, res=" << res, -1);

        ncvAssertReturn(subNoise.size() == subNoiseRecvd.size(), "Data validation failed", -1);
        ncvAssertReturn(0 == memcmp(&subNoise[0], &subNoiseRecvd[0], subNoiseRecvd.size()), "Data validation failed", -1);

        const int numIter = 50/(i+1);
        for (int j=0; j<numIter; j++)
        {
            LIBSSH2_SFTP_HANDLE *sftpHandleDown2;
            AutoSftpHandle aSftpHandleDown(sessionWrap, sftpChannel, "/tmp/test", LIBSSH2_FXF_READ, sftpHandleDown2, SSH_EAGAIN_TIMEOUT_MS, res);
            ncvAssertReturn(res == 0, "AutoSftpHandle ctor failure", -1);

            std::vector<unsigned char> subNoiseRecvd2;
            res = timed_libssh2_sftp_read_all(sessionWrap, sftpChannel, sftpHandleDown2, subNoiseRecvd2, 0, SSH_EAGAIN_TIMEOUT_MS);
            ncvAssertReturn(res == 0, "timed_libssh2_sftp_write_all, res=" << res, -1);
        }

        ncvPrint("Series of " << sz[i] << " bytes downloaded via SFTP");
    }

    return 0;
}
